import uuid
import logging
import json
import threading
from dataclasses import dataclass, field
from typing import Dict

import httpx
# For apiv2
#import awscrt
#import awsiot
#from awsiot import mqtt_connection_builder
from AWSIoTPythonSDK.MQTTLib import AWSIoTMQTTClient
from AWSIoTPythonSDK.exception.AWSIoTExceptions import connectTimeoutException, publishTimeoutException, subscribeTimeoutException
from .model import JciHitachiAWSStatus, JciHitachiAWSStatusSupport

AWS_COGNITO_REGION = "ap-northeast-1"
AWS_COGNITO_IDP_ENDPOINT = f"cognito-idp.{AWS_COGNITO_REGION}.amazonaws.com/"
AWS_COGNITO_ENDPOINT = f"cognito-identity.{AWS_COGNITO_REGION}.amazonaws.com/"
AWS_COGNITO_CLIENT_ID = "7kfnjsb66ei1qt5s5gjv6j1lp6"
AWS_COGNITO_USERPOOL_ID = "ap-northeast-1_aTZeaievK"

AWS_IOT_ENDPOINT = "https://iot-api.jci-hitachi-smarthome.com"
AWS_MQTT_ENDPOINT = "a8kcu267h96in-ats.iot.ap-northeast-1.amazonaws.com"
AWS_POOL_ID = "ap-northeast-1:3bb17635-a6c0-431f-b484-b8de026544cf"
AWS_GATEWAY_KEY = "606153a77f347c9b7becfeda37edb2ccde46ce863c7ac"

_LOGGER = logging.getLogger(__name__)

@dataclass
class AWSTokens:
    access_token: str
    id_token: str
    refresh_token: str

@dataclass
class AWSIdentity:
    identity_id: str
    user_name: str
    user_attributes: dict

@dataclass
class AWSCredentials:
    access_key_id: str
    secret_key: str
    session_token: str

@dataclass
class JciHitachiMqttEvents:
    device_status: Dict[str, JciHitachiAWSStatus] = field(default_factory=dict)
    device_support: Dict[str, JciHitachiAWSStatusSupport] = field(default_factory=dict)
    device_control: Dict[str, dict] = field(default_factory=dict)
    mqtt_error: str = field(default_factory=str)
    device_status_event: threading.Event = field(default_factory=threading.Event)
    device_support_event: threading.Event = field(default_factory=threading.Event)
    device_control_event: threading.Event = field(default_factory=threading.Event)
    mqtt_error_event: threading.Event = field(default_factory=threading.Event)


class JciHitachiAWSCognitoConnection:
    """Connecting to Jci-Hitachi AWS Cognito API.

    Parameters
    ----------
    email : str
        User email.
    password : str
        User password.
    aws_tokens : AWSTokens, optional
        If aws_tokens is given, it is used by request;
        otherwise, a login procedure is performed to obtain new aws_tokens,
        by default None.
    proxy : str, optional
        Proxy setting. Format:"IP:port", by default None. 
    print_response : bool, optional
        If set, all responses of httpx will be printed, by default False.
    """
    
    def __init__(self, email, password, aws_tokens=None, proxy=None, print_response=False):
        self._login_response = None
        self._email = email
        self._password = password
        self._print_response = print_response
        self._proxies = {'http': proxy, 'https': proxy} if proxy else None

        if aws_tokens:
            self._aws_tokens = aws_tokens
        else:
            conn_status, self._aws_tokens = self.login()
            if conn_status != "OK":
                raise RuntimeError(f"An error occurred when signing into AWS Cognito Service: {conn_status}")
    
    def _generate_headers(self, target):
        normal_headers = {
            "X-Amz-Target": target,
            "User-Agent": "Dalvik/2.1.0",
            "content-type": "application/x-amz-json-1.1",
            "Accept" : "application/json",
        }
        return normal_headers

    def _handle_response(self, response):
        response_json = response.json()
        
        if response.status_code == httpx.codes.ok:    
            return "OK", response_json
        else:
            return f"{response_json['__type']} {response_json['message']}", response_json

    @property
    def aws_tokens(self):
        return self._aws_tokens

    def login(self, use_refresh_token=False):
        """Login API.

        Parameters
        ----------
        use_refresh_token : bool, optional
            Whether or not to use AWSTokens.refresh_token to login, by default False

        Returns
        -------
        (str, AWSTokens)
            (status, aws tokens).
        """

        # https://docs.aws.amazon.com/cognito-user-identity-pools/latest/APIReference/API_InitiateAuth.html
        if use_refresh_token and self._aws_tokens != None:
            login_json_data = {
                "AuthFlow": 'REFRESH_TOKEN_AUTH',
                "AuthParameters":{
                    'REFRESH_TOKEN': self._aws_tokens.refresh_token,
                },
                "ClientId": AWS_COGNITO_CLIENT_ID,
            }
        else:
            login_json_data = {
                "AuthFlow": 'USER_PASSWORD_AUTH',
                "AuthParameters": {
                    'USERNAME': self._email,
                    'PASSWORD': self._password,
                },
                "ClientId": AWS_COGNITO_CLIENT_ID,
            }

        login_headers = self._generate_headers("AWSCognitoIdentityProviderService.InitiateAuth")

        login_req = httpx.post("{}".format(f"https://{AWS_COGNITO_IDP_ENDPOINT}"), 
            json=login_json_data,
            headers=login_headers,
            proxies=self._proxies,
        )

        if self._print_response:
            self.print_response(login_req)

        status, response = self._handle_response(login_req)

        aws_tokens = None
        if login_req.status_code == httpx.codes.ok:        
            auth_result = response["AuthenticationResult"]
            aws_tokens = AWSTokens(
                access_token = auth_result['AccessToken'],
                id_token = auth_result['IdToken'],
                refresh_token = auth_result['RefreshToken'],
            )
        return status, aws_tokens

    def get_data(self):
        raise NotImplementedError

    def print_response(self, response):
        print('===================================================')
        print(self.__class__.__name__, 'Response:')
        print('headers:', response.headers)
        print('status_code:', response.status_code)
        print('text:', json.dumps(response.json(), indent=True))
        print('===================================================')


class ChangePassword(JciHitachiAWSCognitoConnection):
    """API internal endpoint. 
    https://docs.aws.amazon.com/cognito-user-identity-pools/latest/APIReference/API_ChangePassword.html
    
    Parameters
    ----------
    email : str
        User email.
    password : str
        User password.
    """

    def __init__(self, email, password, **kwargs):
        super().__init__(email, password, **kwargs)
    
    def get_data(self, new_password):
        json_data = {
            "AccessToken": self._aws_tokens.access_token,
            "PreviousPassword": self._password,
            "ProposedPassword": new_password,
        }

        headers = self._generate_headers("AWSCognitoIdentityProviderService.ChangePassword")

        req = httpx.post("{}".format(f"https://{AWS_COGNITO_IDP_ENDPOINT}"), 
            json=json_data,
            headers=headers,
            proxies=self._proxies,
        )

        if self._print_response:
            self.print_response(req)

        status, response = self._handle_response(req)

        return status, None


class GetUser(JciHitachiAWSCognitoConnection):
    """API internal endpoint. 
    https://docs.aws.amazon.com/cognito-user-identity-pools/latest/APIReference/API_GetUser.html
    
    Parameters
    ----------
    email : str
        User email.
    password : str
        User password.
    """

    def __init__(self, email, password, **kwargs):
        super().__init__(email, password, **kwargs)
    
    def get_data(self):
        json_data = {
            "AccessToken": self._aws_tokens.access_token,
        }

        headers = self._generate_headers("AWSCognitoIdentityProviderService.GetUser")

        req = httpx.post("{}".format(f"https://{AWS_COGNITO_IDP_ENDPOINT}"), 
            json=json_data,
            headers=headers,
            proxies=self._proxies,
        )

        if self._print_response:
            self.print_response(req)

        status, response = self._handle_response(req)

        aws_identity = None
        if req.status_code == httpx.codes.ok:
            user_attributes = {attr["Name"]: attr["Value"] for attr in response["UserAttributes"]}
            aws_identity = AWSIdentity(
                identity_id = user_attributes["custom:cognito_identity_id"],
                user_name = response["Username"],
                user_attributes = user_attributes,
            )
        return status, aws_identity


class GetCredentials(JciHitachiAWSCognitoConnection):
    """API internal endpoint.
    
    Parameters
    ----------
    email : str
        User email.
    password : str
        User password.
    """

    def __init__(self, email, password, **kwargs):
        super().__init__(email, password, **kwargs)

    def get_data(self, aws_identity):
        json_data = {
            "IdentityId": aws_identity.identity_id,
            "Logins": {
                f"{AWS_COGNITO_IDP_ENDPOINT}{AWS_COGNITO_USERPOOL_ID}": self._aws_tokens.id_token,
            }
        }

        headers = self._generate_headers("AWSCognitoIdentityService.GetCredentialsForIdentity")

        req = httpx.post("{}".format(f"https://{AWS_COGNITO_ENDPOINT}"), 
            json=json_data,
            headers=headers,
            proxies=self._proxies,
        )

        if self._print_response:
            self.print_response(req)

        status, response = self._handle_response(req)

        aws_credentials = None
        if req.status_code == httpx.codes.ok:
            aws_credentials = AWSCredentials(
                access_key_id = response['Credentials']['AccessKeyId'],
                secret_key = response['Credentials']['SecretKey'],
                session_token = response['Credentials']['SessionToken'],
            )
        return status, aws_credentials


class JciHitachiAWSIoTConnection:
    """Connecting to Jci-Hitachi AWS IoT API.

    Parameters
    ----------
    aws_tokens : AWSTokens
        AWS tokens.
    proxy : str, optional
        Proxy setting. Format:"IP:port", by default None. 
    print_response : bool, optional
        If set, all responses of httpx will be printed, by default False.
    """

    def __init__(self, aws_tokens, proxy=None, print_response=False):
        self._aws_tokens = aws_tokens
        self._print_response = print_response
        self._proxies = {'http': proxy, 'https': proxy} if proxy else None
    
    def _generate_normal_headers(self):
        normal_headers = {
            "authorization": f"Bearer {self._aws_tokens.id_token}",
            "accesstoken": f"Bearer {self._aws_tokens.access_token}",
            "User-Agent": "Dalvik/2.1.0",
            "content-type" : "application/json",
            "Accept" : "application/json",
        }
        return normal_headers

    def _handle_response(self, response):
        response_json = response.json()
        code = response_json["status"]["code"]
        if code == 0:
            return code, "OK", response_json
        elif code == 6:
            return code, "Invalid email or password", response_json
        elif code == 12:
            return code, "Invalid session token", response_json
        else:
            return code, "Unknown error", response_json

    def _send(self, api_name, json=None):
        req = httpx.post(
            "{}{}".format(AWS_IOT_ENDPOINT, api_name),
            headers=self._generate_normal_headers(),
            json=json,
            proxies=self._proxies,
        )
        if self._print_response:
            self.print_response(req)

        code, message, response_json = self._handle_response(req)

        return message, response_json

    def get_data(self):
        raise NotImplementedError

    def print_response(self, response):
        print('===================================================')
        print(self.__class__.__name__, 'Response:')
        print('headers:', response.headers)
        print('status_code:', response.status_code)
        print('text:', json.dumps(response.json(), indent=True))
        print('===================================================')


class GetAllDevice(JciHitachiAWSIoTConnection):
    """API internal endpoint.
    
    Parameters
    ----------
    aws_tokens : AWSTokens
        AWS tokens.
    """

    def __init__(self, aws_tokens, **kwargs):
        super().__init__(aws_tokens, **kwargs)

    def get_data(self):
        return self._send("/GetAllDevice")


class GetAllGroup(JciHitachiAWSIoTConnection):
    """API internal endpoint.
    
    Parameters
    ----------
    aws_tokens : AWSTokens
        AWS tokens.
    """

    def __init__(self, aws_tokens, **kwargs):
        super().__init__(aws_tokens, **kwargs)

    def get_data(self):
        return self._send("/GetAllGroup")


class GetAllRegion(JciHitachiAWSIoTConnection):
    """API internal endpoint.
    
    Parameters
    ----------
    aws_tokens : AWSTokens
        AWS tokens.
    """

    def __init__(self, aws_tokens, **kwargs):
        super().__init__(aws_tokens, **kwargs)

    def get_data(self):
        return self._send("/GetAllRegion")


class GetAvailableAggregationMonthlyData(JciHitachiAWSIoTConnection):
    """API internal endpoint.
    
    Parameters
    ----------
    aws_tokens : AWSTokens
        AWS tokens.
    """

    def __init__(self, aws_tokens, **kwargs):
        super().__init__(aws_tokens, **kwargs)

    def get_data(self, thing_name, time_start, time_end):
        json_data = {
            "ThingName": thing_name,
            "TimeStart":time_start,
            "TimeEnd":time_end,
        }
        return self._send("/GetAvailableAggregationMonthlyData", json_data)


class GetHistoryEventByUser(JciHitachiAWSIoTConnection):
    """API internal endpoint.
    
    Parameters
    ----------
    aws_tokens : AWSTokens
        AWS tokens.
    """

    def __init__(self, aws_tokens, **kwargs):
        super().__init__(aws_tokens, **kwargs)

    def get_data(self, time_start, time_end):
        json_data = {
            "TimeStart": time_start,
            "TimeEnd": time_end,
        }
        return self._send("/GetHistoryEventByUser", json_data)



class ListSubUser(JciHitachiAWSIoTConnection):
    """API internal endpoint.
    
    Parameters
    ----------
    aws_tokens : AWSTokens
        AWS tokens.
    """

    def __init__(self, aws_tokens, **kwargs):
        super().__init__(aws_tokens, **kwargs)

    def get_data(self):
        return self._send("/ListSubUser")


class JciHitachiAWSMqttConnection:
    """Connecting to Jci-Hitachi AWS MQTT to get latest events. 
        # TODO: Swiching to awscrt API v2 when precompiled wheels of musl linux are available

    Parameters
    ----------
    aws_credentials : AWSCredentials
        AWS credentials.
    print_response : bool, optional
        If set, all responses of MQTT will be printed, by default False.
    """

    def __init__(self, aws_credentials, print_response=False):
        self._aws_credentials = aws_credentials
        self._print_response = print_response
        
        self._mqttc = None
        self._mqtt_events = JciHitachiMqttEvents()

    def __del__(self):
        self.disconnect()
    
    @property
    def mqtt_events(self):
        """MQTT events.

        Returns
        -------
        JciHitachiMqttEvents
            See JciHitachiMqttEvents.
        """

        return self._mqtt_events
    
    def _on_publish(self, client, userdata, message):
        try:
            payload = json.loads(message.payload.decode())
        except Exception as e:
            self._mqtt_events.mqtt_error = e.__class__.__name__
            self._mqtt_events.mqtt_error_event.set()
            _LOGGER.error(e)

        if self._print_response:
            print(f"Mqtt topic {message.topic} published with payload \n {payload}")

        splitted_topic = message.topic.split('/')

        thing_name = splitted_topic[0]

        if len(splitted_topic) >= 3 and splitted_topic[1] == "status" and splitted_topic[2] == "response":
            self._mqtt_events.device_status[thing_name] = JciHitachiAWSStatus(payload)
            self._mqtt_events.device_status_event.set()
        elif len(splitted_topic) >= 3 and splitted_topic[1] == "registration" and splitted_topic[2] == "response":
            self._mqtt_events.device_support[thing_name] = JciHitachiAWSStatusSupport(payload)
            self._mqtt_events.device_support_event.set()
        elif len(splitted_topic) >= 3 and splitted_topic[1] == "control" and splitted_topic[2] == "response":
            self._mqtt_events.device_control[thing_name] = payload
            self._mqtt_events.device_control_event.set()

    def _on_publish_apiv2(self, topic, payload, dup, qos, retain, **kwargs):
        try:
            payload = json.loads(payload.decode())
        except Exception as e:
            self._mqtt_events.mqtt_error = e.__class__.__name__
            self._mqtt_events.mqtt_error_event.set()
            _LOGGER.error(e)

        if self._print_response:
            print(f"Mqtt topic {topic} published with payload \n {payload}")

        splitted_topic = topic.split('/')

        thing_name = splitted_topic[0]

        if len(splitted_topic) >= 3 and splitted_topic[1] == "status" and splitted_topic[2] == "response":
            self._mqtt_events.device_status[thing_name] = JciHitachiAWSStatus(payload)
            self._mqtt_events.device_status_event.set()
        elif len(splitted_topic) >= 3 and splitted_topic[1] == "registration" and splitted_topic[2] == "response":
            self._mqtt_events.device_support[thing_name] = JciHitachiAWSStatusSupport(payload)
            self._mqtt_events.device_support_event.set()
        elif len(splitted_topic) >= 3 and splitted_topic[1] == "control" and splitted_topic[2] == "response":
            self._mqtt_events.device_control[thing_name] = payload
            self._mqtt_events.device_control_event.set()

    def _on_message(self, topic, payload, dup, qos, retain, **kwargs):
        try:
            payload = json.loads(payload.decode())
        except Exception as e:
            self._mqtt_events.mqtt_error = e.__class__.__name__
            self._mqtt_events.mqtt_error_event.set()
            _LOGGER.error(e)
        #if self._print_response:
        #    print(f"{topic} {str(payload)}")
        
        splitted_topic = topic.split('/')

        return
    
    def configure(self):
        """Configure MQTT.
        """
        import os
        self._mqttc = AWSIoTMQTTClient(str(uuid.uuid4()), useWebsocket=True)
        self._mqttc.configureEndpoint(f"{AWS_MQTT_ENDPOINT}", 443)
        # https://github.com/aws/aws-iot-device-sdk-python/issues/273#issuecomment-719897331
        self._mqttc.configureCredentials(os.path.join(os.path.dirname(os.path.abspath(__file__)), './cert/AmazonRootCA1.pem'))
        self._mqttc.configureIAMCredentials(
            self._aws_credentials.access_key_id,
            self._aws_credentials.secret_key,
            self._aws_credentials.session_token
        )
        self._mqttc.configureOfflinePublishQueueing(-1)  # Infinite offline Publish queueing
        self._mqttc.configureDrainingFrequency(10)  # Draining: 2 Hz
        self._mqttc.configureConnectDisconnectTimeout(10)  # 10 sec
        self._mqttc.configureMQTTOperationTimeout(15)  # 5 sec

    def connect(self, topics):
        """Connect to the MQTT broker and start loop.
        
        Parameters
        ----------
        topics : str or list of str
            Topics to subscribe.
        """
        
        try:
            self._mqttc.connect()
        except connectTimeoutException as e:
            _LOGGER.error('Connection timed out.')
        except Exception as e:
            self._mqtt_events.mqtt_error = e.__class__.__name__
            self._mqtt_events.mqtt_error_event.set()
            _LOGGER.error('Connection failed with exception: {}'.format(e))

        if isinstance(topics, str):
            topics = [topics]

        for topic in topics:
            try:
                self._mqttc.subscribe(topic, 1, self._on_publish)
            except subscribeTimeoutException as e:
                _LOGGER.error('Subscription timed out.')
            except Exception as e:
                self._mqtt_events.mqtt_error = e.__class__.__name__
                self._mqtt_events.mqtt_error_event.set()
                _LOGGER.error('Subscription failed with exception: {}'.format(e))

    def publish(self, topic, payload):
        """Publish message.
        
        Parameters
        ----------
        topics : str
            Topics to publish.
        payload : dict
            Payload to publish.
        """

        try:
            self._mqttc.publish(topic, json.dumps(payload), 1)
        except publishTimeoutException as e:
            _LOGGER.error('Publish timed out.')
        except Exception as e:
            self._mqtt_events.mqtt_error = e.__class__.__name__
            self._mqtt_events.mqtt_error_event.set()
            _LOGGER.error('Publish failed with exception: {}'.format(e))

    def disconnect(self):
        """Disconnect from the MQTT broker.
        """

        self._mqttc.disconnect()

    def configure_apiv2(self):
        """Configure MQTT.
        """
        cred_provider = awscrt.auth.AwsCredentialsProvider.new_static(
            self._aws_credentials.access_key_id, 
            self._aws_credentials.secret_key, 
            self._aws_credentials.session_token
        )
        event_loop_group = awscrt.io.EventLoopGroup(1)
        host_resolver = awscrt.io.DefaultHostResolver(event_loop_group)
        client_bootstrap = awscrt.io.ClientBootstrap(event_loop_group, host_resolver)
        self._mqttc = mqtt_connection_builder.websockets_with_default_aws_signing(
            AWS_COGNITO_REGION,
            cred_provider,
            client_bootstrap=client_bootstrap,
            endpoint=AWS_MQTT_ENDPOINT,
            client_id=str(uuid.uuid4())
        )
        self._mqttc.on_message(self._on_message)

    def connect_apiv2(self, topics):
        """Connect to the MQTT broker and start loop.
        
        Parameters
        ----------
        topics : str or list of str
            Topics to subscribe.
        """
        
        try:
            connect_future = self._mqttc.connect()
            print(connect_future.result())
            _LOGGER.info("Connected!")
        except Exception as e:
            self._mqtt_events.mqtt_error = e.__class__.__name__
            self._mqtt_events.mqtt_error_event.set()
            _LOGGER.error('Connection failed with exception {}'.format(e))

        if isinstance(topics, str):
            topics = [topics]
        
        for topic in topics:
            try:
                subscribe_future, _ = self._mqttc.subscribe(topic, awscrt.mqtt.QoS.AT_LEAST_ONCE, callback=self._on_publish)
                print(subscribe_future.result())
            except Exception as e:
                self._mqtt_events.mqtt_error = e.__class__.__name__
                self._mqtt_events.mqtt_error_event.set()
                _LOGGER.error('Subscription failed with exception {}'.format(e))

    def publish_apiv2(self, topic, payload):
        """Publish message.
        
        Parameters
        ----------
        topics : str
            Topics to publish.
        payload : dict
            Payload to publish.
        """

        try:
            publish_future, _ = self._mqttc.publish(topic, json.dumps(payload), awscrt.mqtt.QoS.AT_LEAST_ONCE)
            print(publish_future.result())
        except Exception as e:
            self._mqtt_events.mqtt_error = e.__class__.__name__
            self._mqtt_events.mqtt_error_event.set()
            _LOGGER.error('Publish failed with exception: {}'.format(e))



import random
import time
import warnings
from typing import Dict, List, Optional, Union

from . import aws_connection, connection, mqtt_connection
from .status import (
    JciHitachiCommand,
    JciHitachiCommandAC,
    JciHitachiCommandDH,
    JciHitachiStatusInterpreter,
)
from .model import (
    JciHitachiStatusSupport,
    JciHitachiACSupport,
    JciHitachiDHSupport,
    JciHitachiStatus,
    JciHitachiAC,
    JciHitachiDH,
    JciHitachiAWSStatusSupport,
    JciHitachiAWSStatus,
)


class Peripheral:
    """Peripheral (Device) Information.

    Parameters
    ----------
    peripheral_json : dict
        Peripheral json of specific device.
    """

    supported_device_type = {
        144: "AC",
        146: "DH",
        #148: "HE",
    }

    def __init__(self, peripheral_json : dict) -> None:
        self._json = peripheral_json
        self._available = False
        self._status_code = ""
        self._support_code = ""
        self._supported_status = None

        warnings.warn("This API has been deprecated.", DeprecationWarning)

    def __repr__(self) -> str:
        ret = f"name: {self.name}\n" + \
              f"brand: {self.brand}\n" + \
              f"model: {self.model}\n" + \
              f"type: {self.type}\n" + \
              f"available: {self.available}\n" + \
              f"status_code: {self.status_code}\n" + \
              f"support_code: {self.support_code}\n" + \
              f"gateway_id: {self.gateway_id}\n" + \
              f"gateway_mac_address: {self.gateway_mac_address}"
        return ret

    @classmethod
    def from_device_names(
        cls,
        peripherals_json : dict,
        device_names : Optional[Union[List[str], str]]
    ) -> Dict[str, object]:
        """Use device names to pick peripheral_jsons accordingly.

        Parameters
        ----------
        peripherals_json : dict
            Peripherals_json retrieved from GetPeripheralsByUser.
        device_names : list of str or str or None
            Device name. If None is given, all available devices will be included, by default None.

        Returns
        -------
        dict
            A dict of Peripheral instances with device name key.
        """

        peripherals = {}

        if isinstance(device_names, str):
            device_names = [device_names]

        for result in peripherals_json["results"]:
            device_name = result["DeviceName"]
            device_type = result["Peripherals"][0]["DeviceType"]

            if device_names is None or (device_names and device_name in device_names):
                if device_type in cls.supported_device_type:
                    peripherals[device_name] = cls(result)

        assert device_names is None or len(device_names) == len(peripherals), \
            "Some of device_names are not available from the API."

        return peripherals

    @property
    def available(self) -> bool:
        """Whether the device is available.

        Returns
        -------
        bool
            Return True if the device is available.
        """

        return self._available

    @available.setter
    def available(self, x) -> None:
        self._available = x

    @property
    def brand(self) -> str:
        """Device brand.

        Returns
        -------
        str
            Device brand.
        """
        
        return getattr(self.supported_status, "brand")

    @property
    def commander(self) -> Union[JciHitachiCommand, None]:
        """Return a new JciHitachiCommand instance based on peripheral's type.

        Returns
        -------
        JciHitachiCommand or None
            JciHitachiCommand instance.
        """

        if self.type == "AC":
            return JciHitachiCommandAC(self.gateway_mac_address)
        elif self.type == "DH":
            return JciHitachiCommandDH(self.gateway_mac_address)
        else:
            return None

    @property
    def gateway_id(self) -> int:
        """Gateway ID.

        Returns
        -------
        int
            Gateway ID.
        """

        return self._json["ObjectID"]

    @property
    def gateway_mac_address(self) -> str:
        """Gateway mac address.

        Returns
        -------
        str
            Gateway mac address.
        """

        return self._json["GMACAddress"]

    @property
    def model(self) -> str:
        """Device model.

        Returns
        -------
        str
            Device model.
        """

        return getattr(self.supported_status, "model")

    @property
    def name(self) -> str:
        """Device name.

        Returns
        -------
        str
            Device name.
        """

        return self._json['DeviceName']

    @property
    def picked_peripheral(self) -> dict:
        """Picked peripheral.

        Returns
        -------
        dict
            Picked peripheral.
        """

        return self._json
    
    @property
    def status_code(self) -> str:
        """Peripheral's status code (LValue) reported by the API.

        Returns
        -------
        str
            Status code.
        """

        return self._status_code

    @status_code.setter
    def status_code(self, x : str) -> None:
        self._status_code = x
    
    @property
    def support_code(self) -> str:
        """Peripheral's support code (LValue) reported by the API.

        Returns
        -------
        str
            Status code.
        """

        return self._support_code

    @support_code.setter
    def support_code(self, x : str) -> None:
        self._support_code = x
        if self.type == "AC" :
            self._supported_status = JciHitachiACSupport(
                JciHitachiStatusInterpreter(x).decode_support()
            )
        elif self.type == "DH":
            self._supported_status = JciHitachiDHSupport(
                JciHitachiStatusInterpreter(x).decode_support()
            )

    @property
    def supported_status(self) -> JciHitachiStatusSupport:
        """Peripheral's supported status converted from support_code.

        Returns
        -------
        JciHitachiStatusSupport
            Supported status.
        """

        return self._supported_status

    @property
    def type(self) -> str:
        """Device type.

        Returns
        -------
        str
            Device type. 
            If not supported, 'unknown' will be returned. (currently supports: `AC`, `DH`)
        """

        return self.supported_device_type.get(
            self._json['Peripherals'][0]['DeviceType'],
            'unknown'
        )


class JciHitachiAPI:
    """Jci-Hitachi API.

    Parameters
    ----------
    email : str
        User email.
    password : str
        User password.
    device_names : list of str or str or None, optional
        Device names. If None is given, all available devices will be included, by default None.
    max_retries : int, optional
        Maximum number of retries when setting status, by default 5.
    device_offline_timeout: float, optional
        Device offline timeout, by default 45.0.
    print_response : bool, optional
        If set, all responses of httpx and MQTT will be printed, by default False.
    """

    def __init__(self, 
        email : str,
        password : str,
        device_names : Optional[Union[List[str], str]] = None,
        max_retries : int = 5,
        device_offline_timeout : float = 45.0,
        print_response: bool = False
    ) -> None:
        self.email = email
        self.password = password
        self.device_names = device_names
        self.max_retries = max_retries
        self.device_offline_timeout = device_offline_timeout
        self.print_response = print_response

        self._mqtt = None
        self._device_id = random.randint(1000, 6999)
        self._peripherals = {}
        self._session_token = None
        self._user_id = None
        self._task_id = 0

        warnings.warn("This API has been deprecated.", DeprecationWarning)

    @property
    def peripherals(self) -> Dict[str, Peripheral]:
        """Picked peripherals.

        Returns
        -------
        dict
            A dict of Peripherals.
        """

        return self._peripherals
    
    @property
    def user_id(self) -> Optional[int]:
        """User ID.

        Returns
        -------
        int
            User ID.
        """

        return self._user_id

    @property
    def task_id(self) -> int:
        """Task ID.

        Returns
        -------
        int
            Serial number counted from 0.
        """

        self._task_id += 1
        return self._task_id

    def _sync_peripherals_availablity(self) -> None:
        device_access_time = self._mqtt.mqtt_events.device_access_time
        for peripheral in self._peripherals.values():
            if peripheral.gateway_id in device_access_time and \
                abs(time.time() - device_access_time[peripheral.gateway_id]) < self.device_offline_timeout:
                peripheral.available = True
            else:
                peripheral.available = False

    def login(self) -> None:
        """Login API.

        Raises
        ------
        RuntimeError
            If a login error occurs, RuntimeError will be raised.
        """

        conn = connection.GetPeripheralsByUser(
            self.email,
            self.password,
            print_response=self.print_response)
        conn_status, conn_json = conn.get_data()
        self._session_token = conn.session_token

        if conn_status == "OK":
            if len(conn_json["results"]) != 0:
                self._user_id = conn_json["results"][0]["Owner"]

            # peripherals
            self._peripherals = Peripheral.from_device_names(
                conn_json,
                self.device_names
            )
            self.device_names = list(self._peripherals.keys())

            # mqtt
            self._mqtt = mqtt_connection.JciHitachiMqttConnection(
                self.email,
                self.password,
                self._user_id,
                print_response=self.print_response,
            )
            self._mqtt.configure()
            self._mqtt.connect()

            # status
            self.refresh_status()

        else:
            raise RuntimeError(
                f"An error occurred when API login: {conn_status}.")
    
    def logout(self) -> None:
        """Logout API.
        """
        
        self._mqtt.disconnect()

    def change_password(self, new_password : str) -> None:
        """Change password.

        Parameters
        ----------
        new_password : str
            New password.

        Raises
        ------
        RuntimeError
            If an error occurs, RuntimeError will be raised.
        """

        conn = connection.UpdateUserCredential(
            self.email,
            self.password,
            session_token=self._session_token,
            print_response=self.print_response
        )
        conn_status, conn_json = conn.get_data(new_password)
        self._session_token = conn.session_token

    def get_status(self, device_name: Optional[str] = None) -> Dict[str, JciHitachiStatus]:
        """Get device status after refreshing status.

        Parameters
        ----------
        device_name : str, optional
            Getting a device's status by its name.
            If None is given, all devices' status will be returned,
            by default None.

        Returns
        -------
        dict of JciHitachiStatus.
            Return a dict of JciHitachiStatus instances according to device type.
            For example, if the device type is `AC`, then return JciHitachiAC instance.
        """

        statuses = {}
        for name, peripheral in self._peripherals.items():
            if (device_name and name != device_name) or \
                    peripheral.type == "unknown":
                continue
            dev_status = JciHitachiStatusInterpreter(peripheral.status_code).decode_status()

            if peripheral.type == "AC":
                statuses[name] = JciHitachiAC(dev_status)
            elif peripheral.type == "DH":
                statuses[name] = JciHitachiDH(dev_status)
        return statuses
    
    def get_supported_status(self, device_name: Optional[str] = None) -> Dict[str, JciHitachiStatusSupport]:
        """Get supported device status after refreshing status.

        Parameters
        ----------
        device_name : str, optional
            Getting a device's status by its name.
            If None is given, all devices' status will be returned,
            by default None.

        Returns
        -------
        dict of JciHitachiStatusSupport.
            Return a dict of JciHitachiStatusSupport instances according to device type.
            For example, if the device type is `AC`, then return JciHitachiACSupport instance.
        """

        supported_statuses = {}
        for name, peripheral in self._peripherals.items():
            if (device_name and name != device_name) or \
                    peripheral.type == "unknown":
                continue
            supported_statuses[name] = peripheral.supported_status
        
        return supported_statuses

    def refresh_status(self, device_name : Optional[str] = None) -> None:
        """Refresh device status from the API.

        Parameters
        ----------
        device_name : str, optional
            Refreshing a device's status by its name.
            If None is given, all devices' status will be refreshed,
            by default None.
        
        Raise
        -------
        RuntimeError
            If an error occurs, RuntimeError will be raised.
        """

        self._sync_peripherals_availablity()

        conn = connection.GetDataContainerByID(
            self.email,
            self.password,
            session_token=self._session_token,
            print_response=self.print_response
        )
        for name, peripheral in self._peripherals.items():
            if (device_name and name != device_name) or \
                    peripheral.type == "unknown":
                continue
            conn_status, conn_json = conn.get_data(
                peripheral.picked_peripheral
            )
            self._session_token = conn.session_token

            if conn_status == 'OK':
                peripheral.support_code = conn_json["results"]["DataContainer"][0]["ContDetails"][0]["LValue"]
                peripheral.status_code = conn_json["results"]["DataContainer"][0]["ContDetails"][1]["LValue"]
            else:
                raise RuntimeError(
                    f"An error occurred when refreshing status: {conn_status}"
                )

    def set_status(self, status_name : str, status_value : int, device_name : str) -> bool:
        """Set status to a peripheral.

        Parameters
        ----------
        status_name : str
            Status name, which has to be in idx dict. E.g. JciHitachiAC.idx
        status_value : int
            Status value.
        device_name : str
            Device name.

        Returns
        -------
        bool
            Return True if the command has been successfully executed. Otherwise, return False.

        Raise
        -------
        RuntimeError
            If an error occurs, RuntimeError will be raised.
        """

        self._sync_peripherals_availablity()
        if not self._peripherals[device_name].available:
            return False

        commander = self._peripherals[device_name].commander

        conn = connection.CreateJob(
            self.email,
            self.password,
            session_token=self._session_token,
            print_response=self.print_response
        )
        conn2 = connection.GetJobDoneReport(
            self.email,
            self.password,
            session_token=self._session_token,
            print_response=self.print_response
        )

        # The mqtt events occurring order:
        # job (occurs one time) -> 
        # job_done_report (occurs many times until the job status is successful) ->
        # peripheral (occurs one time, if this step is timed out, the job command fails)
        self._mqtt.mqtt_events.job.clear()
        self._mqtt.mqtt_events.job_done_report.clear()
        self._mqtt.mqtt_events.peripheral.clear()

        conn_status, conn_json = conn.get_data(
            gateway_id=self._peripherals[device_name].gateway_id,
            device_id=self._device_id,
            task_id=self.task_id,
            job_info=commander.get_b64command(status_name, status_value)
        )
        self._session_token = conn.session_token

        if not self._mqtt.mqtt_events.job.wait(timeout=10.0):
            return False
        for _ in range(self.max_retries):
            if not self._mqtt.mqtt_events.job_done_report.wait(timeout=10.0):
                self._mqtt.mqtt_events.job_done_report.clear()
                continue
            conn_status, conn_json = conn2.get_data(
                device_id=self._device_id
            )
            if conn_status == 'OK' and \
                len(conn_json['results']) != 0 and \
                conn_json['results'][0]['JobStatus'] == 0:
                if not self._mqtt.mqtt_events.peripheral.wait(timeout=10.0):
                    return False
                time.sleep(0.5) # still needs to wait a moment.
                return True
        return False


class AWSThing:
    """AWS thing (device) information.

    Parameters
    ----------
    thing_json : dict
        Thing json of specific device.
    """

    supported_device_type = {
        "1": "AC",
        "2": "DH",
        #3: "HE",
        #4: "PM25_PANEL",
    }

    def __init__(self, thing_json : dict) -> None:
        self._json = thing_json
        self._available = True
        self._status_code = None
        self._support_code = None

    def __repr__(self) -> str:
        ret = f"name: {self.name}\n" + \
              f"brand: {self.brand}\n" + \
              f"model: {self.model}\n" + \
              f"type: {self.type}\n" + \
              f"available: {self.available}\n" + \
              f"status_code: {self.status_code}\n" + \
              f"support_code: {self.support_code}\n" + \
              f"gateway_mac_address: {self.gateway_mac_address}"
        return ret

    @classmethod
    def from_device_names(
        cls,
        things_json : dict,
        device_names : Optional[Union[List[str], str]]
    ) -> Dict[str, object]:
        """Use device names to pick peripheral_jsons accordingly.

        Parameters
        ----------
        peripherals_json : dict
            Peripherals_json retrieved from GetPeripheralsByUser.
        device_names : list of str or str or None
            Device name. If None is given, all available devices will be included, by default None.

        Returns
        -------
        dict
            A dict of Peripheral instances with device name key.
        """

        things = {}

        if isinstance(device_names, str):
            device_names = [device_names]

        for thing in things_json["results"]["Things"]:
            device_name = thing["CustomDeviceName"]
            device_type = thing["DeviceType"]

            if device_names is None or (device_names and device_name in device_names):
                if device_type in cls.supported_device_type:
                    things[device_name] = cls(thing)

        assert device_names is None or len(device_names) == len(things), \
            "Some of device_names are not available from the API."

        return things

    @property
    def available(self) -> bool:
        """Whether the device is available.

        Returns
        -------
        bool
            Return True if the device is available.
        """

        return self._available

    @available.setter
    def available(self, x) -> None:
        self._available = x

    @property
    def brand(self) -> str:
        """Device brand.

        Returns
        -------
        str
            Device brand.
        """
        
        return getattr(self._support_code, "Brand")

    @property
    def gateway_mac_address(self) -> str:
        """Gateway mac address.

        Returns
        -------
        str
            Gateway mac address.
        """

        return self._json["ThingName"].split("_")[-1]

    @property
    def model(self) -> str:
        """Device model.

        Returns
        -------
        str
            Device model.
        """

        return getattr(self._support_code, "Model")

    @property
    def name(self) -> str:
        """Device name.

        Returns
        -------
        str
            Device name.
        """

        return self._json['CustomDeviceName']

    @property
    def picked_thing(self) -> dict:
        """Picked thing.

        Returns
        -------
        dict
            Picked thing.
        """

        return self._json
    
    @property
    def status_code(self) -> JciHitachiAWSStatus:
        """Peripheral's status code reported by the API.

        Returns
        -------
        JciHitachiAWSStatus
            Status code.
        """

        return self._status_code

    @status_code.setter
    def status_code(self, x : JciHitachiAWSStatus) -> None:
        self._status_code = x
    
    @property
    def support_code(self) -> JciHitachiAWSStatusSupport:
        """Peripheral's support code reported by the API.

        Returns
        -------
        JciHitachiAWSStatusSupport
            Status code.
        """

        return self._support_code

    @support_code.setter
    def support_code(self, x : JciHitachiAWSStatusSupport) -> None:
        self._support_code = x

    @property
    def supported_status(self) -> JciHitachiStatusSupport:
        """Peripheral's supported status converted from support_code.

        Returns
        -------
        JciHitachiStatusSupport
            Supported status.
        """

        return self._supported_status

    @property
    def thing_name(self):
        return self._json["ThingName"]

    @property
    def type(self) -> str:
        """Device type.

        Returns
        -------
        str
            Device type. 
            If not supported, 'unknown' will be returned. (currently supports: `AC`, `DH`)
        """

        return self.supported_device_type.get(
            self._json['DeviceType'],
            'unknown'
        )


class JciHitachiAWSAPI:
    """Jci-Hitachi API.

    Parameters
    ----------
    email : str
        User email.
    password : str
        User password.
    device_names : list of str or str or None, optional
        Device names. If None is given, all available devices will be included, by default None.
    max_retries : int, optional
        Maximum number of retries when setting status, by default 5.
    device_offline_timeout: float, optional
        For furture use.
    print_response : bool, optional
        If set, all responses of httpx and MQTT will be printed, by default False.
    """

    def __init__(self, 
        email : str,
        password : str,
        device_names : Optional[Union[List[str], str]] = None,
        max_retries : int = 5,
        device_offline_timeout : float = 45.0,
        print_response: bool = False
    ) -> None:
        self.email = email
        self.password = password
        self.device_names = device_names
        self.max_retries = max_retries
        self.device_offline_timeout = device_offline_timeout
        self.print_response = print_response

        self._mqtt = None
        self._device_id = random.randint(1000, 6999)
        self._things = {}
        self._aws_tokens = None
        self._aws_identity = None
        self._aws_credentials = None
        self._host_identity_id = None
        self._error_counter = 0
        self._task_id = 0

    @property
    def things(self) -> Dict[str, AWSThing]:
        """Picked things.

        Returns
        -------
        dict of AWSThing
            A dict of AWSThing.
        """

        return self._things

    @property
    def task_id(self) -> int:
        """Task ID.

        Returns
        -------
        int
            Serial number counted from 0.
        """

        self._task_id += 1
        return self._task_id

    def _check_before_publish(self) -> None:
        if self._mqtt.mqtt_events.mqtt_error_event.is_set() and\
            self._mqtt.mqtt_events.mqtt_error == "wssHandShakeError":
            self._mqtt.mqtt_events.mqtt_error_event.clear()
            self.login()

    def login(self) -> None:
        """Login API.

        Raises
        ------
        RuntimeError
            If a login error occurs, RuntimeError will be raised.
        """

        conn = aws_connection.GetUser(
            email=self.email,
            password=self.password,
            print_response=self.print_response,
        )
        self._aws_tokens = conn.aws_tokens
        conn_status, self._aws_identity = conn.get_data()

        conn = aws_connection.GetCredentials(
            email=self.email,
            password=self.password,
            aws_tokens=self._aws_tokens,
            print_response=self.print_response,
        )
        conn_status, self._aws_credentials = conn.get_data(self._aws_identity)

        conn = aws_connection.ListSubUser(self._aws_tokens, print_response=self.print_response)
        conn_status, conn_json = conn.get_data()

        if conn_status == "OK":
            for user in conn_json["results"]["FamilyMemberList"]:
                if user["isHost"]:
                    self._host_identity_id = user["userId"]
                    break
        else:
            raise RuntimeError(f"An error occurred when listing account users: {conn_status}")

        conn = aws_connection.GetAllDevice(self._aws_tokens, print_response=self.print_response)
        conn_status, conn_json = conn.get_data()

        if conn_status == "OK":
            # Things
            self._things = AWSThing.from_device_names(
                conn_json,
                self.device_names
            )
            self.device_names = list(self._things.keys())

            # mqtt
            self._mqtt = aws_connection.JciHitachiAWSMqttConnection(self._aws_credentials, print_response=self.print_response)
            self._mqtt.configure()

            topics = [f"{self._host_identity_id}_{thing.gateway_mac_address}/#" for thing in self._things.values()]
            self._mqtt.connect(topics=topics)

            # status
            self.refresh_status(refresh_support_code=True)
        else:
            raise RuntimeError(f"An error occurred when retrieving device info: {conn_status}")
    
    def logout(self) -> None:
        """Logout API.
        """
        
        self._mqtt.disconnect()

    def change_password(self, new_password: str) -> None:
        """Change password. 
            Warning: 
            Use this function carefully, be sure you specify a strong enough password; 
            otherwise, your password might be accepted by the Hitachi account management but not be accepted by AWS Cognito, 
            which will result in a login failure in the APP.

        Parameters
        ----------
        new_password : str
            New password.

        Raises
        ------
        RuntimeError
            If an error occurs, RuntimeError will be raised.
        """

        # We have to change both Hitachi owned account management & AWS Cognito simultaneously.
        conn = connection.UpdateUserCredential(
            self.email,
            self.password,
            print_response=self.print_response
        )
        hitachi_conn_status, _ = conn.get_data(new_password)

        conn = aws_connection.ChangePassword(self.email, self.password, aws_tokens=self._aws_tokens, print_response=self.print_response)
        aws_conn_status, _ = conn.get_data(new_password)

        if hitachi_conn_status != "OK" or aws_conn_status != "OK":
            raise RuntimeError(f"An error occurred when changing password: {hitachi_conn_status} {aws_conn_status}")

    def refresh_status(self, device_name : Optional[str] = None, refresh_support_code=False) -> None:
        """Refresh device status from the API.

        Parameters
        ----------
        device_name : str, optional
            Refreshing a device's status by its name.
            If None is given, all devices' status will be refreshed,
            by default None.
        refresh_support_code : bool, optional
            Whether or not to refresh support code.
        
        Raise
        -------
        RuntimeError
            If an error occurs, RuntimeError will be raised.
        """

        self._check_before_publish()

        for name, thing in self._things.items():
            if (device_name and name != device_name) or \
                    thing.type == "unknown":
                continue

            if refresh_support_code:
                self._mqtt.publish(f"{self._host_identity_id}_{thing.gateway_mac_address}/registration/request", {"Timestamp": time.time()})
                if not self._mqtt.mqtt_events.device_support_event.wait(timeout=10.0):
                    raise RuntimeError(
                        f"An error occurred when refreshing support code."
                    )
                else:
                    thing.support_code = self._mqtt.mqtt_events.device_support.get(thing.thing_name)
                self._mqtt.mqtt_events.device_support_event.clear()
                time.sleep(0.5)

            self._mqtt.publish(f"{self._host_identity_id}_{thing.gateway_mac_address}/status/request", {"Timestamp": time.time()})
            if not self._mqtt.mqtt_events.device_status_event.wait(timeout=10.0):
                raise RuntimeError(
                    f"An error occurred when refreshing status code."
                )
            else:
                thing.status_code = self._mqtt.mqtt_events.device_status.get(thing.thing_name)
            self._mqtt.mqtt_events.device_status_event.clear()
            
    
    def get_status(self, device_name: Optional[str] = None, legacy=False) -> Dict[str, JciHitachiAWSStatus]:
        """Get device status after refreshing status.

        Parameters
        ----------
        device_name : str, optional
            Getting a device's status by its name.
            If None is given, all devices' status will be returned,
            by default None.
        legacy : bool, optional
            Whether or not to return legacy (old) status class.

        Returns
        -------
        dict of JciHitachiAWSStatus or JciHitachiStatus.
            if legacy is True, return a dict of JciHitachiAWSStatus; otherwise, return a dict of JciHitachiAWSStatus instances.
        """
        
        statuses = {}
        for name, thing in self._things.items():
            if (device_name and name != device_name) or \
                    thing.type == "unknown":
                continue
            if legacy:
                statuses[name] = thing.status_code.legacy_status_class
            else:
                statuses[name] = thing.status_code
        return statuses

    def set_status(self, status_name: str, status_value: int, device_name: str) -> bool:
        """Set status to a thing.

        Parameters
        ----------
        status_name : str
            Status name.
        status_value : int
            Status value.
        device_name : str
            Device name.

        Returns
        -------
        bool
            Return True if the command has been successfully executed. Otherwise, return False.

        Raise
        -------
        RuntimeError
            If an error occurs, RuntimeError will be raised.
        """

        self._check_before_publish()
        #if not self._peripherals[device_name].available:
        #    return False

        thing_name = self._things[device_name].thing_name
        thing_type = self._things[device_name].type
        gateway_mac_address = self._things[device_name].gateway_mac_address
        if status_name not in JciHitachiAWSStatus.compability_mapping[thing_type]:
            status_name = JciHitachiAWSStatus.convert_old_to_new(thing_type, status_name)


        # prevent publishing status when a device status is refreshing
        #self._mqtt.mqtt_events.device_status_event.wait(timeout=5.0)
        #self._mqtt.mqtt_events.device_support_event.wait(timeout=5.0)

        self._mqtt.publish(f"{self._host_identity_id}_{gateway_mac_address}/control/request", {
            "Condition": {
                "ThingName": thing_name,
                "Index": 0,
                "Geofencing": {
                    "Arrive": None,
                    "Leave": None,
                },
            },
            status_name: status_value,
            "TaskID": self.task_id,
            "Timestamp": time.time(),
        })

        for _ in range(self.max_retries):
            if not self._mqtt.mqtt_events.device_control_event.wait(timeout=10.0):
                device_control = None
            else:
                device_control = self._mqtt.mqtt_events.device_control.get(thing_name)
                if device_control.get(status_name) == status_value:
                    self._mqtt.mqtt_events.device_control_event.clear()
                    time.sleep(0.5)
                    return True
            time.sleep(0.5)
        return False


import json
import logging
import os
import ssl
import threading
from dataclasses import dataclass, field
from typing import Dict

import paho.mqtt.client as mqtt

from .utility import convert_hash

MQTT_ENDPOINT = "mqtt.jci-hitachi-smarthome.com"
MQTT_PORT = 8893
MQTT_VERSION = 4
MQTT_SSL_CERT = os.path.join(os.path.dirname(os.path.abspath(__file__)), "cert/mqtt-jci-hitachi-smarthome-com-chain.pem")
MQTT_SSL_CONTEXT = ssl.create_default_context(cafile=MQTT_SSL_CERT)
MQTT_SSL_CONTEXT.set_ciphers("DEFAULT@SECLEVEL=1")  # the cert uses SHA1-RSA1024bits ciphers so unfortunately we have to lower the security level
MQTT_SSL_CONTEXT.hostname_checks_common_name = True  # the cert lacks a subjectaltname

_LOGGER = logging.getLogger(__name__)


@dataclass
class JciHitachiMqttEvents:
    device_access_time: Dict[str, int] = field(default_factory=dict)
    job: threading.Event = field(default_factory=threading.Event)
    job_done_report: threading.Event = field(default_factory=threading.Event)
    peripheral: threading.Event = field(default_factory=threading.Event)


class JciHitachiMqttConnection:
    """Connecting to Jci-Hitachi MQTT to get latest events.

    Parameters
    ----------
    email : str
        User email.
    password : str
        User password.
    user_id : int
        User ID.
    print_response : bool, optional
        If set, all responses of MQTT will be printed, by default False.
    """

    def __init__(self, email, password, user_id, print_response=False):
        self._email = email
        self._password = password
        self._user_id = user_id
        self._print_response = print_response
        
        self._mqttc = mqtt.Client()
        self._mqtt_events = JciHitachiMqttEvents()

    def __del__(self):
        self.disconnect()

    @property
    def mqtt_events(self):
        """MQTT events.

        Returns
        -------
        JciHitachiMqttEvents
            See JciHitachiMqttEvents.
        """

        return self._mqtt_events

    def _on_connect(self, client, userdata, flags, rc):
        if self._print_response:
            print(f"Mqtt connected with result code {rc}")

        client.subscribe(f"out/ugroup/{self._user_id}/#")
    
    def _on_disconnect(self, client, userdata, rc):
        if self._print_response:
            print(f"Mqtt disconnected with result code {rc}")
        
        if rc == mqtt.MQTT_ERR_SUCCESS:
            self._mqttc.loop_stop()
        else:
            _LOGGER.error(f"Unexpected MQTT disconnection: {mqtt.error_string(rc)}.")

    def _on_message(self, client, userdata, msg):
        if self._print_response:
            print(f"{msg.topic} {str(msg.payload)}")
        
        splitted_topic = msg.topic.split('/')
        payload = json.loads(msg.payload)
        if len(splitted_topic) == 6 and splitted_topic[-1] == "status":
            gateway_id = payload["args"]["ObjectID"]
            time = payload["time"]
            self._mqtt_events.device_access_time[gateway_id] = time
        elif len(splitted_topic) == 4 and splitted_topic[-1] == "resp":
            if payload["args"]["Name"] == "JobDoneReport":
                self._mqtt_events.job_done_report.set()
            elif payload["args"]["Name"] == "Peripheral":
                self._mqtt_events.peripheral.set()
        elif len(splitted_topic) == 4 and splitted_topic[-1] == "job":
            if payload["args"]["Name"] == "Job":
                self._mqtt_events.job.set()

    def configure(self):
        """Configure MQTT.
        """

        self._mqttc.username_pw_set(f"$MAIL${self._email}", f"{self._email}{convert_hash(f'{self._email}{self._password}')}")
        self._mqttc.tls_set_context(MQTT_SSL_CONTEXT)
        self._mqttc.on_connect = self._on_connect
        self._mqttc.on_disconnect = self._on_disconnect
        self._mqttc.on_message = self._on_message

        if self._print_response:
            self._mqttc.enable_logger(logger=_LOGGER)

    def connect(self):
        """Connect to the MQTT broker and start loop.
        """

        self._mqttc.connect_async(MQTT_ENDPOINT, port=MQTT_PORT, keepalive=60, bind_address="")
        self._mqttc.loop_start()

    def disconnect(self):
        """Disconnect from the MQTT broker.
        """

        self._mqttc.disconnect()